## -*- coding: utf-8 -*-

__author__ = "gael@pilotsystems.net"
__format__ = "plaintext"
__version__ = "$Id: SharedBaseTool.py 3 2007-11-14 16:18:37Z gael.le-mignot $"

from Products.CMFCore.utils import getToolByName
from workflowgraph import MultiWorkflowGraph
from App.config import getConfiguration

import xmlrpclib, os, base64

import logging
log = logging.getLogger('SharedDocumentBase')


class Handler(object):
    """
    Abstract class for handler
    """
    cases = []

    def __init__(self, tool, event):
        """
        Constructor
        """
        self.tool = tool
        self.portal = getToolByName(tool, "portal_url").getPortalObject()
        self.event = event

    def resolve_folder(self):
        """
        Return folder containing the object and its id
        """
        path = self.event.path.split("/")
        folder = "/".join(path[:-1])
        try:
            folder = self.portal.unrestrictedTraverse(folder)
        except:
            self.warning("can't traverse to container")
            return None, None
        objid = path[-1]
        return folder, objid

    def resolve_object(self):
        """
        Resolve event path locally
        """
        try:
            return self.portal.unrestrictedTraverse(self.event.path)
        except:
            self.warning("can't traverse to container")
            return None

    def resolve_remote(self):
        """
        Resolve event path on target
        """
        mapper, session = self.tool.getSQLMapper()        
        plone = mapper.getMapper("plone")
        plones = session.query(plone).filter(plone.c.id == self.event.plone)
        if plones.count() > 1:
            raise RuntimeError, "Duplicate plone id %r" % self.event.plone
        if plones.count() == 0:
            raise KeyError, "No such plone id %r" % self.event.plone
        plone = plones[0]
        url = plone.url.split("/")
        path = self.event.path.split("/")
        url = url + path
        # Inject username and password
        username, password = self.tool.getRemoteUser()
        url[2] = "%s:%s@%s" % (username, password, url[2])
        url = "/".join(url)
        # Contact it
        target = xmlrpclib.Server(url)
        return target

    def warning(self, message):
        """
        Display formatted warning
        """
        log.warning("%s (%r): %s" % (self.__class__.__name__,
                                     self.event.path, message))
    
    def info(self, message):
        """
        Display formatted info
        """
        log.info("%s (%r): %s" % (self.__class__.__name__,
                                  self.event.path, message))
    
    def process(self):
        """
        The main process method
        """
        raise NotImplementedError

class DelHandler(Handler):
    """
    Handle object suppression
    """
    cases = [ ("del", "*") ]

    def process(self):
        """
        The main process method
        """
        folder, objid = self.resolve_folder()
        if not folder:
            return
        try:
            folder.manage_delObjects([ objid ])
        except KeyError:
            self.warning("object didn't exist")

class FolderAddedHandler(Handler):
    """
    Handle folder creation
    """
    cases = [ ("add", "SharedFolder") ]

    def process(self):
        """
        The main process method
        """
        folder, objid = self.resolve_folder()
        if not folder:
            return
        folder.invokeFactory("SharedFolder", objid)
        # Delegate attributes loading to edition hanlder
        handler = get_handler("edit", "SharedFolder")    
        handler(self.tool, self.event).process()        

class FolderEditedHandler(Handler):
    """
    Handle folder edition
    """
    cases = [ ("edit", "SharedFolder") ]

    def process(self):
        """
        The main process method
        """
        obj = self.resolve_object()
        if not obj:
            return
        remote = self.resolve_remote()
        obj.unmarkCreationFlag()
        obj.edit(title = remote.Title(),
                 description = remote.Description())

        # Handle workflow state
        review_state = remote.get_review_state()        
        wfgraph = MultiWorkflowGraph(self.portal)
        wfgraph.set_to(obj, review_state)

class FileEditedHandler(Handler):
    """
    Handle file edition and creation
    """
    cases = [ ("edit", "SharedFile"),
              ("add", "SharedFile"), ]

    def process(self):
        """
        The main process method
        """
        folder, objid = self.resolve_folder()
        if not folder:
            return
        remote = self.resolve_remote()

        # Emit warning if mode is not wanted mode
        if hasattr(folder.aq_explicit, objid):
            mode = "edit"
        else:
            mode = "add"
        if self.event.type != mode:
            self.warning("we need to %r and we were asked to %r" % (mode,
                                                                    self.event.type))
        # Edit ? Well, delete it
        if mode == "edit":
            folder.manage_delObjects([ objid ])

        # Import data
        data = base64.decodestring(remote.get_zexp_data())
        cfg = getConfiguration()
        zexpname = objid + ".zexp"
        importPath = os.path.join(cfg.instancehome, 'import', zexpname)
        importfile = open(importPath, "w")
        importfile.write(data)
        importfile.close()
        try:
            folder.manage_importObject(zexpname)
        finally:
            os.remove(importPath)
            
        # Handle workflow state
        obj = getattr(folder.aq_explicit, objid)
        review_state = remote.get_review_state()        
        wfgraph = MultiWorkflowGraph(self.portal)
        wfgraph.set_to(obj, review_state)

    
allhandlers = [ DelHandler, FolderAddedHandler, FolderEditedHandler,
                FileEditedHandler ]
handlers = {}
for handler in allhandlers:
    for case in handler.cases:
        handlers[case] = handler

def get_handler(event_type, portal_type):
    """
    Get handler matching for us
    """
    attempts = [ (event_type, portal_type),
                 (event_type, "*"), ("*", portal_type), ("*", "*") ]
    for attempt in attempts:
        if handlers.has_key(attempt):
            return handlers[attempt]
    return None
