## -*- coding: utf-8 -*-

__author__ = "gael@pilotsystems.net"
__format__ = "plaintext"
__version__ = "$Id: SharedBaseTool.py 8 2007-11-29 15:57:25Z gael.le-mignot $"

from Globals import InitializeClass
from OFS.SimpleItem import SimpleItem
from AccessControl import ClassSecurityInfo, Unauthorized
from OFS.PropertyManager import PropertyManager

from Products.CMFCore.utils import UniqueObject

from Products.CMFCore.utils import getToolByName

from z3c.sqlalchemy import createSAWrapper, getSAWrapper

from Products.SharedDocumentBase.ActionHandlers import get_handler

import md5

import logging
log = logging.getLogger('SharedDocumentBase')

import threading, transaction, thread
from Products.SharedDocumentBase.utils import run_as_user

from Products.SharedDocumentBase.SharedQueue import SharedQueue

_global_shared_mutex = threading.RLock()
_global_shared_queue = SharedQueue()

META_TYPE = "SharedBaseTool"
TOOL_ID = "portal_shared_document_base"

class SharedBaseTool(UniqueObject, SimpleItem, PropertyManager):
    """
    Shared document base tool.
    """
    id = TOOL_ID
    title = "Shared document base Tool"
    meta_type = META_TYPE

    manage_options = (PropertyManager.manage_options + 
                      ({ 'label'  : 'Initial setup',
                         'action' : 'manage_initialSetupForm',
                         },) + 
                      ({ 'label'  : 'Replay script',
                         'action' : 'manage_replayScriptForm',
                         },) + 
                      SimpleItem.manage_options)

    ## Define properties
    _properties = ({'id': 'database_url',
                    'label': 'Database connection string',
                    'mode': 'w',
                    'type': 'string'},
                   {'id': 'current_id',
                    'label': 'Current script position',
                    'mode': 'w',
                    'type': 'int' },
                   )

    ## Set default values for properties
    database_url = "mysql://user:password@host/database"
    current_id = 0

    ## Security 
    security = ClassSecurityInfo()


    security.declareProtected("Manage portal", 'manage_initialSetup')
    def manage_initialSetup(self):
        """
        Wrapper for ZMI
        """
        self.setupDatabase()
        return self.REQUEST.RESPONSE.redirect(self.absolute_url() + '/manage_initialSetupForm?portal_status_message=Setup+done.')

    security.declareProtected("Manage portal", 'manage_replayScript')
    def manage_replayScript(self):
        """
        Wrapper for ZMI
        In ZMI, we do synchronize with update thread
        """
        done = self.replayScript()
        if done:
            msg = "Script+replayed."
        else:
            msg = "Script+replay+scheduled."
        return self.REQUEST.RESPONSE.redirect(self.absolute_url() + '/manage_replayScriptForm?portal_status_message=' + msg)

    security.declarePrivate("getSQLMapper")
    def getSQLMapper(self):
        """
        Get current SQL Mapper, or create a new one
        """
        name = "/".join(self.getPhysicalPath())
        try:
            mapper = getSAWrapper(name)
        except ValueError:
            log.info("Creating new SQL mapper")
            mapper = createSAWrapper(self.database_url, forZope=True, name=name)
        session = mapper.session
        return mapper, session

    security.declarePublic("getUid")
    def getUid(self):
        """
        Compute a unique Id for this Plone site
        """
        portal = getToolByName(self, "portal_url").getPortalObject()
        path = tuple(__file__.split("/")[:-4]) + portal.getPhysicalPath()
        path = "/".join(path)
        date = portal.CreationDate()
        uid = path + "::" + str(date)
        return md5.new(uid).hexdigest()

    security.declarePublic("getPloneId")
    def getPloneId(self):
        """
        Get our Plone id
        """
        uid = self.getUid()
        mapper, session = self.getSQLMapper()
        plone = mapper.getMapper("plone")
        res = session.query(plone).filter(plone.c.uid == uid)

        if res.count() > 1: # More than one entry ? panic !
            raise RuntimeError, "Duplicate UID %r" % uid
        if res.count() == 1: # One entry ? fine
            return res[0].id

        return None

    security.declarePrivate("getRemoteUser")
    def getRemoteUser(self):
        """
        Get the remote username and password
        """
        mapper, session = self.getSQLMapper()
        config = mapper.getMapper("config")
        res = session.query(config).filter(config.c.key == "username")
        if res.count() == 0:
            raise RuntimeError, "Missing configuration for username"
        username = res[0].value
        res = session.query(config).filter(config.c.key == "password")
        if res.count() == 0:
            raise RuntimeError, "Missing configuration for password"
        password = res[0].value
        return username, password

    security.declareProtected("Manage portal", "setupDatabase")
    def setupDatabase(self, check = True):
        """
        Will ensure we have the Plone site referenced into the database,
        and create the Manager user if needed
        """
        uid = self.getUid()
        mapper, session = self.getSQLMapper()

        # Handle plone site creation
        plone = mapper.getMapper("plone")
        create = not check
        if check:
            if not self.getPloneId():
                create = True
        if create:
            portal = getToolByName(self, "portal_url").getPortalObject()
            entry = plone()
            entry.uid = uid
            entry.url = portal.absolute_url()
            log.info("Creating new Plone entry for id %r" % uid)
            session.save(entry)
            
        # Handle user creation
        username, password = self.getRemoteUser()
        try:
            self.acl_users.source_users.manage_addUser(username, username, 
                                                       password, password)
            self.acl_users.portal_role_manager.assignRoleToPrincipal('Manager',
                                                                     username)
            log.info("User %r created." % username)
        except KeyError:
            log.info("User %r already existed." % username)
            

    security.declarePublic("getPath")
    def getPath(self, obj):
        """
        Get the relative path of an object
        """
        portal = getToolByName(self, "portal_url").getPortalObject()
        path = obj.getPhysicalPath()
        path = path[len(portal.getPhysicalPath()):]
        return "/".join(path)

    security.declareProtected("View", "recordEntry")
    def recordEntry(self, action, obj):
        """
        Record new entry
        """
        txn = transaction.get()
        if getattr(txn, "_doing_script_replay", False):
            # Not registring entries while we are replaying
            return

        pid = self.getPloneId()
        if not pid:
            log.warning("We have so SQL id, please run the setup")
            return
        mapper, session = self.getSQLMapper()
        events = mapper.getMapper("events")
        event = events()
        event.plone = pid
        event.path = self.getPath(obj)
        event.portal_type = obj.portal_type
        event.type = action
        session.save(event)

    security.declareProtected("Manage portal", "mergeEvents")
    def mergeEvents(self, old, new):
        """
        Merge two events into a new one
        Return None if the merge resolves into nothing
        """
        mapper, session = self.getSQLMapper()
        events = mapper.getMapper("events")

        prefix = "Object on path %r (%s, %s): " %  (old.path, old.id, new.id)

        if old.type == "add":
            if new.type == "add":
                log.warning(prefix + "is added twice")
                return old
            if new.type == "edit":
                log.info(prefix + "merging edit after add")
                event = events()
                event.path = new.path
                event.plone = new.plone
                event.type = "add"
                event.portal_type = new.portal_type
                return event
            if new.type == "del":
                log.info(prefix + "merging del after add")
                return None

        if old.type == "edit":
            if new.type == "add":
                log.warning(prefix + "is added after edition")
                return new
            if new.type == "edit":
                log.info(prefix + "merging edit after edit")
                return new
            if new.type == "del":
                log.info(prefix + "merging del after edit")
                return new
                
        if old.type == "del":
            if new.type == "add":
                log.info(prefix + "merging addition after deletion")
                return new
            if new.type == "edit":
                log.warning(prefix + "is edited after deletion")
                return new
            if new.type == "del":
                log.warning(prefix + "is deleted twice")
                return old
                
        return None

    security.declareProtected("Manage portal", "loadScript")
    def loadScript(self):
        """
        Load current script
        This will compact entries too
        """
        entries = {}
        ploneid = self.getPloneId()
        mapper, session = self.getSQLMapper()        
        events = mapper.getMapper("events")

        # First, load events
        evts = session.query(events).filter(events.c.id > self.current_id)
        max_id = self.current_id
        for event in evts:
            max_id = max(event.id, max_id)
            if event.plone == self.getPloneId():
                continue
            if not entries.has_key(event.path):
                entries[event.path] = event
                continue
            # Ok, we already have something for the same path
            # Let's do some optimization
            old = entries[event.path]
            new = self.mergeEvents(old, event)
            if new is None:
                del entries[event.path]
            else:
                entries[new.path] = new
        self.current_id = max_id

        # Then, order things by path
        entries = entries.values()
        entries.sort(lambda e1, e2: cmp(e1.path, e2.path))

        # Now, do some filtering
        # If we have a "del" entry, we shouldn't process any other event below
        # this level
        deleted = None
        result = []
        for entry in entries:
            if deleted:
                # We already have a deleted item
                if entry.path.startswith(deleted.path):
                    if entry.type != "del":
                        log.warning("We should have %sed %r inside the deleted %r" % (entry.type, entry.path, deleted.path))
                    continue
                # Out of deletion path
                deleted = None

            # Ok, now we add the entry
            result.append(entry)
            if entry.type == "del":
                deleted = entry

        return result

    security.declareProtected("Manage portal", "replayScript")
    def replayScript(self):
        """
        Replay all new entries added in the database since last run

        Ok, here is the catch: we can't do it naively, or we'll have deadlocks
        when all threads try to replay the script at once.

        So, if one is already doing the replay, we'll queue ourself, and it'll
        take care of our work when it finished its own work
        """

        _global_shared_queue.append(self.getPhysicalPath())
            
        locked = _global_shared_mutex.acquire(blocking = False)
        if locked:
            try:
                root = self.getPhysicalRoot()
                while True:
                    path = _global_shared_queue.pop()
                    if not path:
                        break
                    tool = root.unrestrictedTraverse("/".join(path))
                    tool.do_replayScript()
            finally:
                _global_shared_mutex.release()

        return locked

    
    security.declarePrivate("do_replayScript")
    def do_replayScript(self):
        """    
        Replay all new entries added in the database since last run
        We synchronize our mutex to prevent any deadlock on querying
        informations to target Zope.
        """
        # That's a bit dirty, but need to be sure we don't record to the
        # database the changes we are doing during script replay
        txn = transaction.get()
        txn._doing_script_replay = True
        try:
            username, password = self.getRemoteUser()
            robot = self.acl_users.aq_inner.source_users.getUser(username)
            entries = self.loadScript()
            for entry in entries:
                handler = get_handler(entry.type, entry.portal_type)
                if not handler:
                    log.warning("No handler for %r" % key)
                else:
                    hdlr = handler(self, entry)
                    run_as_user(robot, hdlr.process)
        finally:
            txn._doing_script_replay = False
        txn.commit()

InitializeClass(SharedBaseTool)
